if RequiredScript == "lib/managers/menumanager" then

	--define internal mod information such as directories
	_G.FasterEffectQueues = {
		mod_path = ModPath,
		loc_path = ModPath .. "loc/",
		save_path = SavePath .. "FasterEffectQueues.txt",
		default_loc_path = ModPath .. "loc/en.txt",
		options_path = ModPath .. "menu/options.txt",
		settings = {
			fbiq_optimizations = true,
			queue_enabled = true,
			max_queue_items = 3
		}
	}

	--getter for max queue size number
	function FasterEffectQueues:GetMaxQueueItems()
		return self.settings.max_queue_items
	end

	--getter for queue enabled bool
	function FasterEffectQueues:IsQueueEnabled()
		return self.settings.queue_enabled
	end
	
	--getter for general optimizations to misc gameplaycentralmanager functions made by hoxi
	function FasterEffectQueues:IsEggCodeEnabled()
		return self.settings.fbiq_optimizations
	end

	--io load from user settings txt
	function FasterEffectQueues:Load()
		local file = io.open(FasterEffectQueues.save_path, "r")

		if file then
			for k, v in pairs(json.decode(file:read("*all"))) do
				FasterEffectQueues.settings[k] = v
			end
		else
			FasterEffectQueues:Save()
		end
	end

	--io save to user settings txt
	function FasterEffectQueues:Save()
		local file = io.open(FasterEffectQueues.save_path,"w+")

		if file then
			file:write(json.encode(FasterEffectQueues.settings))
			file:close()
		end
	end

	local max_steps_per_frame = 3 --in case an option to change this gets added
	local max_hits_per_frame = FasterEffectQueues:GetMaxQueueItems()
	local function set_max_hits_per_frame(n)
		--write to the local var, in order to avoid a getter call for the number from settings every call
		max_hits_per_frame = n
	end

	local function set_queue_enabled(on)

		--if managers.game_play_central isn't yet instantiated, just change the class function instead;
		--else, change the instance function
		local gpcm = managers and managers.game_play_central or GamePlayCentralManager

		if gpcm then
			--change the reference to the function itself, in order to avoid a getter call for the "enabled" bool from settings every call
			if on then
				gpcm._flush_bullet_hits = FasterEffectQueues._flush_bullet_hits__framelocked
				gpcm.play_impact_sound_and_effects = FasterEffectQueues.play_impact_sound_and_effects__framelocked
			else
				gpcm._flush_bullet_hits = FasterEffectQueues._flush_bullet_hits__frameless
				gpcm.play_impact_sound_and_effects = FasterEffectQueues.play_impact_sound_and_effects__frameless
			end
		--else
--			log("FasterEffectQueues: GamePlayCentralManager not found")
		end
	end

	Hooks:Add("LocalizationManagerPostInit", "LocalizationManagerPostInit_feq", function( loc )
		--you don't need this if you're using beardlib to handle localization
		loc:load_localization_file( FasterEffectQueues.default_loc_path)
	end)

	--add the menu callbacks for when menu options are changed
	Hooks:Add( "MenuManagerInitialize", "MenuManagerInitialize_feq", function(menu_manager)
		
		MenuCallbackHandler.callback_feq_fbiq_optimizations = function(self,item)
			local on = item:value() == "on"
			FasterEffectQueues.settings.fbiq_optimizations = on
			
			FasterEffectQueues:Save()
		end
		
		--called when changing the queue enabled toggle
		MenuCallbackHandler.callback_feq_set_queue_enabled = function(self,item)
			local on = item:value() == "on"
			FasterEffectQueues.settings.queue_enabled = on

			set_queue_enabled(on)

			FasterEffectQueues:Save()
		end

		--called when changing the max queue count slider
		MenuCallbackHandler.callback_feq_set_max_queue_items = function(self,item)
			local value = tonumber(item:value())
			--value is a float, so if you wish to apply a math.round, floor, ceil, etc. to turn it into an integer, you should do so here
			FasterEffectQueues.settings.max_queue_items = math.round(value)
			FasterEffectQueues:Save()
		end

		--called when the menu is closed
		MenuCallbackHandler.callback_feq_close = function(self)
			set_max_hits_per_frame(FasterEffectQueues:GetMaxQueueItems())
		end

		--load settings from user's mod settings txt
		FasterEffectQueues:Load()

		--apply max queue count from settings
		set_max_hits_per_frame(FasterEffectQueues:GetMaxQueueItems())

		--apply function overloading according to settings
		set_queue_enabled(FasterEffectQueues:IsQueueEnabled())

		--create menus
		MenuHelper:LoadFromJsonFile(FasterEffectQueues.options_path, FasterEffectQueues, FasterEffectQueues.settings)
	end)


	--do the stuff;
	--functions are saved to the FasterEffectQueues global instead,
	--and the override is applied later according to settings
	--intentionally written without implicit self arg because these should NOT actually be called from FasterEffectQueues

	local mvec3_set = mvector3.set
	local mvec3_set_stat = mvector3.set_static
	local mvec3_mul = mvector3.multiply
	local mvec3_add = mvector3.add
	local mvec3_neg = mvector3.negate
	local mvec3_lerp = mvector3.lerp
	local mvec3_spread = mvector3.spread
	local mvec3_dist_sq = mvector3.distance_sq
	local mvec3_dot = mvector3.dot
	local mvec3_cross = mvector3.cross
	local mvec3_cpy = mvector3.copy
	local tmp_vec1 = Vector3()
	local tmp_vec2 = Vector3()
	local tmp_vec3 = Vector3()
	local tmp_vec4 = Vector3()
	local tmp_vec5 = Vector3()
	local zero_vector = Vector3()

	local zero_rot = Rotation()

	local math_random = math.random
	local math_up = math.UP
	local math_lerp = math.lerp
	local math_clamp = math.clamp
	local math_sign = math.sign

	local next_g = next
	local unpack_g = unpack
	local world_g = World
	local alive_g = alive

	local idstr_func = Idstring
	local empty_idstr = idstr_func("")
	local idstr_blood_spatter = idstr_func("blood_spatter")
	local idstr_blood_screen = idstr_func("effects/particles/character/player/blood_screen")
	local idstr_bullet_hit_blood = idstr_func("effects/payday2/particles/impacts/blood/blood_impact_a")
	local idstr_fallback = idstr_func("effects/payday2/particles/impacts/fallback_impact_pd2")
	local idstr_no_material = idstr_func("no_material")
	local idstr_bullet_hit = idstr_func("bullet_hit")
	local idstr_shotgun_push_effect = idstr_func("physic_effects/shotgun_hit")

	function FasterEffectQueues.play_impact_sound_and_effects__framelocked(self, params)
		if params.immediate then
			self:_play_bullet_hit(params)
		else
			local hits = self._bullet_hits

			hits[#hits + 1] = params

			self._bullet_hits = hits
		end
	end

	function FasterEffectQueues.play_impact_sound_and_effects__frameless(self, params)
		self:_play_bullet_hit(params)
	end

	function FasterEffectQueues._flush_bullet_hits__framelocked(self)
		local bullet_hits = self._bullet_hits
		local nr_hits = #bullet_hits

		if nr_hits > 0 then
			local play_f = self._play_bullet_hit
			local needs_rearranging = nr_hits > max_hits_per_frame
			local to_i = needs_rearranging and max_hits_per_frame or nr_hits

			for i = 1, to_i do
				play_f(self, bullet_hits[i])
			end

			--dont use table.remove, table.remove with an index is really fucking slow
			--remake the table with everything except the indexes we've played, this is legitimately faster than using table.remove
			if needs_rearranging then
				local new_table = {}

				for i = max_hits_per_frame + 1, nr_hits do
					new_table[#new_table + 1] = bullet_hits[i]
				end

				self._bullet_hits = new_table
			else
				self._bullet_hits = {}
			end
		end
	end

	--this is here in case framelocked hits are being used
	--and then the game is paused + the setting gets changed to frameless + effects were still queued
	function FasterEffectQueues._flush_bullet_hits__frameless(self)
		local bullet_hits = self._bullet_hits
		local nr_hits = #bullet_hits

		if nr_hits > 0 then
			local play_f = self._play_bullet_hit

			for i = 1, nr_hits do
				play_f(self, bullet_hits[i])
			end

			self._bullet_hits = {}
		end
	end

	--optimize other functions to remove table.remove/unoptimized functions, completely empty the table after we're done

	function FasterEffectQueues._flush_play_effects(self)
		local play_effects = self._play_effects
		local nr_effects = #play_effects

		if nr_effects > 0 then
			local effect_manager = self._effect_manager
			local spawn_f = effect_manager.spawn

			for i = 1, nr_effects do
				spawn_f(effect_manager, play_effects[i])
			end

			self._play_effects = {}
		end
	end

	function FasterEffectQueues._flush_play_sounds(self)
		local play_sounds = self._play_sounds
		local nr_sounds = #play_sounds

		if nr_sounds > 0 then
			local play_f = self._play_sound

			for i = 1, nr_sounds do
				play_f(self, play_sounds[i])
			end

			self._play_sounds = {}
		end
	end

	function FasterEffectQueues.request_play_footstep(self, unit, m_pos)
		local camera_pos = self._camera_pos

		if camera_pos then
			local footsteps = self._footsteps
			local nr_footsteps = #footsteps
			local dis = mvec3_dist_sq(camera_pos, m_pos)

			if dis < 2250000 then
				if nr_footsteps < max_steps_per_frame then
					self._footsteps[nr_footsteps + 1] = {unit = unit, dis = dis}
				else
					local insert_i, longest_dis = nil

					for i = 1, nr_footsteps do
						local footstep = footsteps[i]

						if not alive_g(footstep.unit) then
							insert_i = i

							break
						else
							local footstep_dis = footstep.dis

							if dis < footstep_dis then
								if not longest_dis or longest_dis < footstep_dis then
									insert_i = i
									longest_dis = footstep_dis
								end
							end
						end
					end

					if insert_i then
						local new_table = {}

						for i = 1, insert_i - 1 do
							new_table[#new_table + 1] = footsteps[i]
						end

						new_table[insert_i] = {unit = unit, dis = dis}

						for i = insert_i + 1, nr_footsteps do
							new_table[#new_table + 1] = footsteps[i]
						end

						self._footsteps = new_table
					end
				end
			end
		end
	end

	function FasterEffectQueues._flush_footsteps(self)
		local footsteps = self._footsteps
		local nr_footsteps = #footsteps

		if nr_footsteps > 0 then
			local slotmask = self._slotmask_footstep
			local material_name_f = self.material_name
			local decal_ray_from = tmp_vec1
			local decal_ray_to = tmp_vec2

			for i = 1, nr_footsteps do
				local footstep = footsteps[i]
				local unit = footstep.unit

				if alive_g(unit) then
					local ext_movement = unit:movement()

					mvec3_set(decal_ray_from, ext_movement:m_head_pos())
					mvec3_set(decal_ray_to, math_up)
					mvec3_mul(decal_ray_to, -250)
					mvec3_add(decal_ray_to, decal_ray_from)

					local mat_name = nil
					local ground_ray_unit = ext_movement:ground_ray()
					ground_ray_unit = ground_ray_unit and ground_ray_unit.unit

					if alive_g(ground_ray_unit) then
						mat_name = world_g:pick_decal_material(ground_ray_unit, decal_ray_from, decal_ray_to, slotmask)
					else
						mat_name = world_g:pick_decal_material(decal_ray_from, decal_ray_to, slotmask)
					end

					local sound_switch_name = mat_name ~= empty_idstr and mat_name or idstr_no_material
					local sound_source = unit:sound_source()
					sound_source:set_switch("materials", material_name_f(self, sound_switch_name))

					local event = ext_movement:get_footstep_event()
					sound_source:post_event(event)
				end
			end

			self._footsteps = {}
		end
	end

	function FasterEffectQueues._play_bullet_hit(self, params)
		local col_ray = params.col_ray
		local hit_unit = col_ray.unit

		if not alive_g(hit_unit) then
			return
		end

		local hit_pos = col_ray.position
		local in_view_f = world_g.in_view_with_options
		local need_sound = not params.no_sound and in_view_f(world_g, hit_pos, 4000, 0, 0)
		local need_effect = in_view_f(world_g, hit_pos, 20, 100, 5000)
		local need_decal = not self._block_bullet_decals and not params.no_decal and need_effect and in_view_f(world_g, hit_pos, 3000, 0, 0)

		if not need_sound and not need_effect and not need_decal then
			return
		end

		local event = params.event or "bullet_hit"
		local decal = params.decal and idstr_func(params.decal) or idstr_bullet_hit
		local slot_mask = params.slot_mask or self._slotmask_bullet_impact_targets

		local hit_dir = col_ray.ray
		local hit_normal = col_ray.normal
		local decal_ray_from = tmp_vec1
		local decal_ray_to = tmp_vec2

		mvec3_set(decal_ray_from, hit_dir)
		mvec3_mul(decal_ray_from, 25)
		mvec3_set(decal_ray_to, hit_pos)
		mvec3_add(decal_ray_to, decal_ray_from)
		mvec3_neg(decal_ray_from)
		mvec3_add(decal_ray_from, hit_pos)

		local material_name = world_g:pick_decal_material(hit_unit, decal_ray_from, decal_ray_to, slot_mask)
		material_name = material_name ~= empty_idstr and material_name
		local effect = params.effect
		local sound_switch_name = nil

		if material_name then
			local redir_name, offset_hit_pos = nil
			local sphere_ray_radius = col_ray.sphere_cast_radius

			if sphere_ray_radius then
				offset_hit_pos = tmp_vec5
				mvec3_set(offset_hit_pos, hit_dir)
				mvec3_mul(offset_hit_pos, sphere_ray_radius)
				mvec3_add(offset_hit_pos, hit_pos)
			end

			if need_decal then
				if offset_hit_pos then
					redir_name = world_g:project_decal(decal, offset_hit_pos, hit_dir, hit_unit, math_up, hit_normal)
				else
					redir_name = world_g:project_decal(decal, hit_pos, hit_dir, hit_unit, math_up, hit_normal)
				end
			elseif need_effect then
				redir_name = world_g:pick_decal_effect(decal, hit_unit, decal_ray_from, decal_ray_to, slot_mask)
			end

			if need_effect then
				redir_name = redir_name == empty_idstr and idstr_fallback or redir_name

				local inv_dir = tmp_vec3
				mvec3_set(inv_dir, hit_dir)
				mvec3_neg(inv_dir)

				local effect_normal = tmp_vec4
				mvec3_lerp(effect_normal, hit_normal, inv_dir, math_random())
				mvec3_spread(effect_normal, 10)

				if offset_hit_pos then
					effect = {
						effect = effect or redir_name,
						position = mvec3_cpy(offset_hit_pos),
						normal = mvec3_cpy(effect_normal)
					}
				else
					effect = {
						effect = effect or redir_name,
						position = hit_pos,
						normal = mvec3_cpy(effect_normal)
					}
				end
			end

			if need_sound then
				sound_switch_name = params.sound_switch_name or material_name
			end
		else
			if need_effect then
				local inv_dir = tmp_vec3
				mvec3_set(inv_dir, hit_dir)
				mvec3_neg(inv_dir)

				local effect_normal = tmp_vec4
				mvec3_lerp(effect_normal, hit_normal, inv_dir, math_random())
				mvec3_spread(effect_normal, 10)

				local generic_effect = effect or idstr_fallback
				effect = {
					effect = generic_effect,
					position = hit_pos,
					normal = mvec3_cpy(effect_normal)
				}
			end

			if need_sound then
				sound_switch_name = params.sound_switch_name or idstr_no_material
			end
		end

		--removal of table.insert shit

		if effect and effect.effect then
			local effects = self._play_effects

			effects[#effects + 1] = effect

			self._play_effects = effects
		end

		if need_sound then
			local sounds = self._play_sounds

			sounds[#sounds + 1] = {sound_switch_name = sound_switch_name, position = hit_pos, event = event}

			self._play_sounds = sounds
		end
	end

	function FasterEffectQueues._get_impact_source(self)
		local impact_sounds = self._impact_sounds
		local cur_index = impact_sounds.index
		local source = impact_sounds.sources[cur_index]

		impact_sounds.index = cur_index < impact_sounds.max_index and cur_index + 1 or 1

		self._impact_sounds = impact_sounds

		return source
	end

	function FasterEffectQueues.play_impact_flesh(self, params)
		local col_ray = params.col_ray
		local hit_unit = col_ray.unit

		if not alive_g(hit_unit) or not hit_unit:in_slot(self._slotmask_flesh) then
			return
		end

		local hit_pos = col_ray.position
		local splatter_to = tmp_vec1
		mvec3_set(splatter_to, col_ray.ray)
		mvec3_mul(splatter_to, 1000)
		mvec3_add(splatter_to, hit_pos)

		local splatter_ray = hit_unit:raycast("ray", hit_pos, splatter_to, "slot_mask", self._slotmask_world_geometry)

		if splatter_ray then
			world_g:project_decal(idstr_blood_spatter, splatter_ray.position, splatter_ray.ray, splatter_ray.unit, nil, splatter_ray.normal)
		end

		local player = managers.player:player_unit()

		if player and mvec3_dist_sq(hit_pos, player:movement():m_head_pos()) < 40000 then
			self._effect_manager:spawn({
				effect = idstr_blood_screen,
				position = zero_vector,
				rotation = zero_rot
			})
		end
	end

	function FasterEffectQueues.sync_play_impact_flesh(self, hit_pos, dir)
		local splatter_to = tmp_vec1
		mvec3_set(splatter_to, dir)
		mvec3_mul(splatter_to, 1000)
		mvec3_add(splatter_to, hit_pos)

		local splatter_ray = world_g:raycast("ray", hit_pos, splatter_to, "slot_mask", self._slotmask_world_geometry)

		if splatter_ray then
			world_g:project_decal(idstr_blood_spatter, splatter_ray.position, splatter_ray.ray, splatter_ray.unit, nil, splatter_ray.normal)
		end

		local player = managers.player:player_unit()

		if player and mvec3_dist_sq(hit_pos, player:movement():m_head_pos()) < 40000 then
			local effect_manager = self._effect_manager
			local spawn_f = effect_manager.spawn

			spawn_f(effect_manager, {
				effect = idstr_bullet_hit_blood,
				position = hit_pos,
				normal = dir
			})
			spawn_f(effect_manager, {
				effect = idstr_blood_screen,
				position = zero_vector,
				rotation = zero_rot
			})
		else
			self._effect_manager:spawn({
				effect = idstr_bullet_hit_blood,
				position = hit_pos,
				normal = dir
			})
		end

		local sound_source = self:_get_impact_source()

		sound_source:stop()
		sound_source:set_position(hit_pos)
		sound_source:set_switch("materials", "flesh")
		sound_source:post_event("bullet_hit")
	end

	--fixes a vanilla crash when a map element enables flashlights on enemies (if they have one set up at all)
	--and there happens to be a grenade/projectile/throwable (since some might be using slot 13, and they all lack a flashlight_state_changed function
	function FasterEffectQueues.set_flashlights_on(self, state)
		if self._flashlights_on == state then
			return
		end

		self._flashlights_on = state

		local weapons = world_g:find_units_quick("all", 13)

		for i = 1, #weapons do
			local weap_base = weapons[i]:base()

			if weap_base and weap_base.flashlight_state_changed then
				weap_base:flashlight_state_changed()
			end
		end
	end

elseif RequiredScript == "lib/managers/gameplaycentralmanager" then
	FasterEffectQueues:Load() --load things again to make sure restart-only options get respected
	local FBIQ = FasterEffectQueues
	--apply the function changes here

	GamePlayCentralManager._get_impact_source = FBIQ._get_impact_source
	GamePlayCentralManager.play_impact_flesh = FBIQ.play_impact_flesh
	GamePlayCentralManager.sync_play_impact_flesh = FBIQ.sync_play_impact_flesh
	GamePlayCentralManager._flush_play_effects = FBIQ._flush_play_effects
	GamePlayCentralManager._flush_play_sounds = FBIQ._flush_play_sounds
	GamePlayCentralManager.request_play_footstep = FBIQ.request_play_footstep
	GamePlayCentralManager._flush_footsteps = FBIQ._flush_footsteps
	GamePlayCentralManager._play_bullet_hit = FBIQ._play_bullet_hit
	GamePlayCentralManager.set_flashlights_on = FBIQ.set_flashlights_on

	--this value may reflect the default value instead of the one in settings, if called before the MenuManagerInitialize hook (which does seem to be the case),
	--since it would then be called before settings are read from the user's settings in the saves folder;
	--so in said hook, we re-apply the function changes, to make sure the function used is the one that properly reflects the user's setting
	local frame_locked = FBIQ:IsQueueEnabled()

	if frame_locked then
		GamePlayCentralManager._flush_bullet_hits = FBIQ._flush_bullet_hits__framelocked
		GamePlayCentralManager.play_impact_sound_and_effects = FBIQ.play_impact_sound_and_effects__framelocked
	else
		GamePlayCentralManager._flush_bullet_hits = FBIQ._flush_bullet_hits__frameless
		GamePlayCentralManager.play_impact_sound_and_effects = FBIQ.play_impact_sound_and_effects__frameless
	end
	
	--hoxi's optimizations require a restart as im not 100% sure of their stability when switched on and off on the fly
	--plus, she said she'd prefer it this way specifically because of stability reasons
	if not FBIQ:IsEggCodeEnabled() then 
		return
	end

	local mvec3_set = mvector3.set
	local mvec3_set_stat = mvector3.set_static
	local mvec3_mul = mvector3.multiply
	local mvec3_add = mvector3.add
	local mvec3_neg = mvector3.negate
	local mvec3_lerp = mvector3.lerp
	local mvec3_spread = mvector3.spread
	local mvec3_dist_sq = mvector3.distance_sq
	local mvec3_dot = mvector3.dot
	local mvec3_cross = mvector3.cross
	local mvec3_cpy = mvector3.copy
	local tmp_vec1 = Vector3()
	local tmp_vec2 = Vector3()
	local tmp_vec3 = Vector3()
	local tmp_vec4 = Vector3()
	local tmp_vec5 = Vector3()
	local zero_vector = Vector3()

	local zero_rot = Rotation()

	local math_random = math.random
	local math_up = math.UP
	local math_lerp = math.lerp
	local math_clamp = math.clamp
	local math_sign = math.sign

	local next_g = next
	local unpack_g = unpack
	local world_g = World
	local alive_g = alive

	local idstr_func = Idstring
	local empty_idstr = idstr_func("")
	local idstr_blood_spatter = idstr_func("blood_spatter")
	local idstr_blood_screen = idstr_func("effects/particles/character/player/blood_screen")
	local default_projectile_trail = idstr_func("effects/payday2/particles/weapons/arrow_trail")
	local idstr_bullet_hit_blood = idstr_func("effects/payday2/particles/impacts/blood/blood_impact_a")
	local idstr_fallback = idstr_func("effects/payday2/particles/impacts/fallback_impact_pd2")
	local idstr_no_material = idstr_func("no_material")
	local idstr_bullet_hit = idstr_func("bullet_hit")
	local idstr_shotgun_push_effect = idstr_func("physic_effects/shotgun_hit")

	--these optional optimizations literally change nothing, it's the same functionality, just optimized, but they might not fit the mod's purpose and or could be mod unfriendly, be sure to clarify this to some extent

	function GamePlayCentralManager.physics_push(self, col_ray, push_multiplier)
		local hit_unit = col_ray.unit

		if not hit_unit:in_slot(self._slotmask_physics_push) then
			return
		end

		local body = col_ray.body

		if not body:dynamic() then
			local original_body_com = body:center_of_mass()
			local closest_body_dis_sq = nil
			local nr_bodies = hit_unit:num_bodies()
			local get_body_f = hit_unit.body
			local i_body = 0

			while nr_bodies > i_body do
				local test_body = get_body_f(hit_unit, i_body)

				if test_body:enabled() and test_body:dynamic() then
					local test_dis_sq = mvec3_dist_sq(test_body:center_of_mass(), original_body_com)

					if not closest_body_dis_sq or test_dis_sq < closest_body_dis_sq then
						closest_body_dis_sq = test_dis_sq
						body = test_body
					end
				end

				i_body = i_body + 1
			end

			--no dynamic body was found, don't do anything
			if not closest_body_dis_sq then
				return
			end
		end

		push_multiplier = push_multiplier or 1

		local body_mass = body:mass()
		body_mass = body_mass > 50 and 50 or body_mass
		body_mass = body_mass * push_multiplier
		local body_vel = body:velocity()
		local push_vec = tmp_vec1

		mvec3_set(push_vec, col_ray.ray)

		local vel_dot = mvec3_dot(body_vel, push_vec)
		local max_vel = 600

		if vel_dot < max_vel then
			vel_dot = vel_dot < 0 and 0 or vel_dot
			local push_vel = max_vel - vel_dot
			push_vel = math_lerp(push_vel * 0.7, push_vel, math_random()) * push_multiplier

			mvec3_mul(push_vec, push_vel)
			body:push_at(body_mass, push_vec, col_ray.position)
		end
	end

	function GamePlayCentralManager._flush_queue_fire_raycast(self)
		local queued_rays = self._queue_fire_raycast
		local nr_rays = #queued_rays

		if nr_rays > 0 then
			local ray_data = queued_rays[1]
			local data = ray_data.data
			local user_unit = data[1]

			if alive_g(user_unit) then
				local weap_unit = ray_data.weapon_unit

				if alive_g(weap_unit) then
					weap_unit:base():_fire_raycast(unpack_g(data))
				end
			end

			if nr_rays == 1 then
				self._queue_fire_raycast = {}
			else
				local new_table = {}

				for i = 2, nr_rays do
					new_table[#new_table + 1] = queued_rays[i]
				end

				self._queue_fire_raycast = new_table
			end
		end
	end

	function GamePlayCentralManager._do_shotgun_push(self, unit, hit_pos, dir, distance, attacker)
		local full_body_action = unit:movement()._active_actions[1]

		if full_body_action and full_body_action:type() == "hurt" then
			full_body_action:force_ragdoll(true)
		end

		local scale = math_clamp(1 - distance / self:get_shotgun_push_range(attacker), 0.5, 1)
		local rot_time = 1 + math_random() * 2
		local base_ext = unit:base()

		--Dozers will get pushed with less force and won't spin around so much
		if base_ext and base_ext.has_tag and base_ext:has_tag("tank") then
			scale = scale * 0.3
			rot_time = rot_time * 0.4
		end

		local push_vec = tmp_vec1
		mvec3_set_stat(push_vec, dir.x, dir.y, dir.z + 0.5)
		mvec3_mul(push_vec, 600 * scale)

		local height_sign = math_sign(mvec3_dist_sq(hit_pos, unit:position()) - 10000)
		local twist_dir = math_random(2) == 1 and 1 or -1
		local rot_acc = tmp_vec2
		mvec3_set(rot_acc, dir)
		mvec3_cross(rot_acc, rot_acc, math_up)
		mvec3_set(tmp_vec3, math_up)
		mvec3_mul(tmp_vec3, 0.5 * twist_dir)
		mvec3_add(rot_acc, tmp_vec3)
		mvec3_mul(rot_acc, -1000 * height_sign)

		local play_physic_effect_f = world_g.play_physic_effect
		local nr_u_bodies = unit:num_bodies()
		local get_body_f = unit.body
		local i_u_body = 0

		while nr_u_bodies > i_u_body do
			local u_body = get_body_f(unit, i_u_body)

			if u_body:enabled() and u_body:dynamic() then
				local body_mass = u_body:mass() * 4 / math_random(2)

				play_physic_effect_f(world_g, idstr_shotgun_push_effect, u_body, push_vec, body_mass, rot_acc, rot_time)
				
				managers.mutators:notify(Message.OnShotgunPush, unit, hit_pos, dir, distance, attacker)
			end

			i_u_body = i_u_body + 1
		end	
	end

	function GamePlayCentralManager.add_projectile_trail(self, unit, object, effect)
		local trails = self._projectile_trails
		local u_key = unit:key()

		if trails[u_key] then
			return
		end

		local effect_ids = effect and idstr_func(effect) or default_projectile_trail
		local effect = self._effect_manager:spawn({
			effect = effect_ids,
			parent = object
		})
		trails[u_key] = {effect = effect}

		self._projectile_trails = trails
	end

	function GamePlayCentralManager.remove_projectile_trail(self, unit)
		local u_key = unit:key()
		local data = self._projectile_trails[u_key]

		if not data or data.total_t then
			return
		end

		data.total_t = 0.5
		data.t = data.total_t
		data.u_key = u_key

		local trails_to_remove = self._projectile_trails_remove or {}
		trails_to_remove[#trails_to_remove + 1] = data
		self._projectile_trails_remove = trails_to_remove
	end

	function GamePlayCentralManager_update_projectile_trails(self, t, dt)
		local trails_to_remove = self._projectile_trails_remove

		if not trails_to_remove then
			return
		end

		local original_trails_table = self._projectile_trails
		local nr_trails = #trails_to_remove
		local effect_manager = self._effect_manager
		local kill_f = effect_manager.kill
		local set_var_float_f = effect_manager.set_simulator_var_float

		for i = nr_trails, 1, -1 do
			local data = trails_to_remove[i]
			local trail_t = data.t - dt

			if trail_t <= 0 then
				--set_var_float_f(effect_manager, data.effect, atom_ids, simulator_ids, opacity_ids, 0)

				kill_f(effect_manager, data.effect)

				if nr_trails == 1 then
					self._projectile_trails_remove = nil
				else
					local new_table = {}

					for idx = 1, i - 1 do
						new_table[#new_table + 1] = trails_to_remove[idx]
					end

					for idx = i + 1, nr_trails do
						new_table[#new_table + 1] = trails_to_remove[idx]
					end

					nr_trails = nr_trails - 1
					trails_to_remove = new_table
					self._projectile_trails_remove = trails_to_remove
				end

				original_trails_table[data.u_key] = nil

				if next_g(original_trails_table) then
					self._projectile_trails = original_trails_table
				else
					self._projectile_trails = {}
				end
			else
				data.t = trail_t

				local opacity = trail_t / data.total_t
				set_var_float_f(effect_manager, data.effect, atom_ids, simulator_ids, opacity_ids, opacity)
			end
		end
	end
	
end
